import os
from cognito.table import Table
from prettytable import PrettyTable


class Report:
    
    def __init__(self, filename):
        self.data = Table(filename)



    def generate(self):

        columns = self.data.columns()
        table = PrettyTable(columns)
        print(table)
        







# -*- coding: utf-8 -*-
import os
import cognito
import pandas as pd
import numpy as np
from cognito.table import Table




def get_stories(df):
    """
    Get intutive stories from given dataset
    
    :param      df:   { parameter_description }
    :type       df:   { type_description }
    """
    categories = df.get_categorical().columns
    continuous = df.get_numerical().columns

    stories = []
    cat_copy = list(categories)
    for col in categories:
        # Remove the current col
        if col in cat_copy:
            cat_copy.remove(col)
        try:
            # Get comparison variable
            x = cat_copy.pop()
            d = pd.pivot_table(df.data, index=(col), values=[x],\
             aggfunc='count').reset_index().sort_values(by=x, ascending=False)
            stories.append({
                'question': "%s with high count of %s" %(col, x),
                'question_html': "<span class='tag is-primary is-light'>%s</span>\
                 with high count of  <span class='tag is-success is-light'>%s</span>" % (col, x),
                'answer': d[col].head(1).values[0],
                'misc': d
            })
        except IndexError as e:
            pass
        
        for num in continuous:
            d = pd.pivot_table(df.data, index=[col], values=[num],\
             aggfunc=np.sum).reset_index().sort_values(by=num, ascending=False)
            stories.append({
                'question': "%s with sum of %s" % (col, num),
                'question_html': "<span class='tag is-primary is-light'>%s</span>\
                with sum of <span class='tag is-success is-light'>%s</span>" % (col, num),
                'answer': round(d[num].head(1).values[0]),
                'misc': d
            })

    return stories


def get_interesting_stories(df):
    """
    Get intutive stories from given dataset
    
    :param      df:   { parameter_description }
    :type       df:   { type_description }
    """
    categories = df.get_categorical().columns
    continuous = df.get_numerical().columns

    stories = []
    cat_copy = list(categories)
    for col in categories:
        # Remove the current col
        if col in cat_copy:
            cat_copy.remove(col)
        try:
            # Get comparison variable
            x = cat_copy.pop()
            d = pd.pivot_table(df.data, index=(col), values=[x],\
             aggfunc='count').reset_index().sort_values(by=x, ascending=False)


            # Highest %s was %s on the %s %s
            stories.append({
                'question': "Highest %s was %s of %s %s." %(x, d[x].head(1).values[0], col, d[col].head(1).values[0]),
                'question_html': "Highest <span class='tag is-primary is-light'>%s</span>\
                 was %s of  <span class='tag is-success is-light'>%s</span> %s." % (x, d[x].head(1).values[0], col, d[col].head(1).values[0]),
                'answer': d[col].head(1).values[0],
                'misc': d
            })

            # Lowest %s was %s on the %s %s
            stories.append({
                'question': "Lowest %s was %s of %s %s." %(x, d[x].tail(1).values[0], col, d[col].tail(1).values[0]),
                'question_html': "Lowest <span class='tag is-primary is-light'>%s</span>\
                 was %s of  <span class='tag is-success is-light'>%s</span> %s." % (x, d[x].tail(1).values[0], col, d[col].tail(1).values[0]),
                'answer': d[col].tail(1).values[0],
                'misc': d
            })


        except IndexError as e:
            pass
        
        for num in continuous:
            d = pd.pivot_table(df.data, index=[col], values=[num],\
             aggfunc=np.sum).reset_index().sort_values(by=num, ascending=False)


            stories.append({
                'question': "Highest %s is %s for %s %s." % (num, d[num].head(1).values[0], col, d[col].head(1).values[0]),
                'question_html': "Highest <span class='tag is-primary is-light'>%s</span>\
                is %s for <span class='tag is-success is-light'>%s</span> %s." % (num, d[num].head(1).values[0], col, d[col].head(1).values[0]),
                'answer': round(d[num].head(1).values[0]),
                'misc': d
            })


            stories.append({
                'question': "Lowest %s is %s for %s %s." % (num, d[num].tail(1).values[0], col, d[col].tail(1).values[0]),
                'question_html': "Lowest <span class='tag is-primary is-light'>%s</span>\
                is %s for <span class='tag is-success is-light'>%s</span> %s." % (num, d[num].tail(1).values[0], col, d[col].tail(1).values[0]),
                'answer': round(d[num].tail(1).values[0]),
                'misc': d
            })

    return stories






"""
Data checking module
"""
# pylint: disable=C0301
from __future__ import print_function
import os
import sys
import math
import re
from datetime import datetime
from os import path
import pandas as pd
import numpy as np
import datefinder

class Check():
    """
    Check class helps us to identify the types of
    variables categorical | Continuous | Discrete
    """
    def __init__(self):
        pass

    @staticmethod
    def is_working(column="Cognito!"):
        """
        Determines whether the specified command is working.
        :param      column:  The column
        :type       column:  string
        """
        return "Hello, %s! How are you %s?"%(column, column)


    @staticmethod
    def is_categorical(column):
        """
        Determines whether the specified col is categorical.
        :param      col:  column name
        :type       col:  { pandas.series | list | tuple }
        :return     boolean: True | False
        
        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/
        
        Usage:
        ======
            >> Check.is_categorical(data['Age'])
            >> True
        """
        try:
            return bool(True) if column.dtypes == 'object' else bool(False)
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def is_continuous(column):
        """
        Determines whether the specified col is continuous.
        :param      col:  column name
        :type       col:  { pandas.series | list | tuple }
        :return     boolean: True | False
        
        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/
        
        Usage:
        ======
            >> Check.is_continuous(data['Age'])
            >> False
        """
        try:
            return bool(True) if column.dtypes == 'float64' else bool(False)
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def is_discrete(column):
        """
        Determines whether the specified column is discrete.
        :param      column:  column name
        :type       column:  { pandas.series | list | tuple }
        :return     boolean: True | False
        
        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/
        
        Usage:
        ======
            >> Check.is_discrete(data['Age'])
            >> True
        """
        try:
            return bool(True) if column.dtypes == 'int64' else bool(False)
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def is_identifier(column):
        """
        Determines whether the specified column is identifier.
        :param      column:  The column
        :type       column:  { pandas.series | list | tuple }
        :return     boolean: True | False
        
        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/

        Usage:
        ======
            >> Check.is_identifier(data['Age'])
            >> True
        """
        try:
            return bool(True) if column.nunique() == column.shape[0] else bool(False)
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def is_missing(column):
        """
        Determines whether the specified column has missing values.

      
        :param      column:  The column
        :type       column:  { pandas.series | list | tuple }
        :return     boolean: True | False
        
        weblink:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4
        
        Usage:
        ======
            >> Check.is_missing(data['population'])
            >> True
        """
        try:
            return bool(True) if column.isnull().values.any() == bool(True) else bool(False)
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def ignore_identifier(dataframe):
        """
        Drops the table if the column is an identifier.
        :param      column:  The column
        :type       column:  { pandas.series | list | tuple }
        :return     DataFrame:Updated DataFrame
       
        Weblink:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4
       
        Usage:
        ======
            >> Check.ignore_identifier(data)
            >> Updated Dataframe
        """
        col = list(dataframe)
        for i in col:
            if Check.is_identifier(dataframe[i]):
                dataframe.drop([i], axis=1, inplace=True)
        return dataframe


    @staticmethod
    def is_outlier(column, threshold):
        """
        Returns all the outliers in the column.

        :param      column:   The column
        :type       column:   {pandas.series | list | tuple}
        :return     list:     List of all the ouliers in the column
       
        Weblink: https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4
       
        Usage:
        =====
           >> Check.is_outlier(data['population'])
           >> [6815.0, 6860.0, 11551.0]
        """
        column = pd.Series(column)
        outliers = []
        mean = np.mean(column)
        std_dev = np.std(column)
        for value in column:
            z_score = (value - mean) / std_dev
            if np.abs(z_score) > threshold:
                outliers.append(value)
        return outliers


    @staticmethod
    def percentage_missing(dataframe):
        """
        Calculates the percentage of missing value in each column of dataframe.

        :param       dataframe:  The dataframe
        :type        dataframe:  { pandas.dataframe }
        :return      dictionary:  Dictionary of column name with percentage of missing values
       
        Weblink: https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4
       
        Usage:
        ======
        >> Check.percentage_missing(data)
        >> {Price:0.00, Age:10.00}

        """
        missing = dataframe.isnull().sum()
        row_size = len(dataframe.index)
        missing_dict = {}
        for i in list(dataframe.columns):
            perc = round(missing[i] / row_size * 100.0, 2)
            missing_dict.update({i:perc})
        return missing_dict


    @staticmethod
    def remove_columns(dataframe):
        """
        Removes the column containing 60% or more missing data.

        :param      dataframe:  The dataframe
        :type       dataframe:  { pandas.dataframe }
        :return     dataframe:  Dataframe with the columns dropped
        
        Weblink:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4
        
        Usage:
        ======
        >>Check.remove_columns(data)
        >>dataframe

        """
        missing_dict = Check.percentage_missing(dataframe)
        for column in missing_dict:
            if missing_dict[column] >= 60.00:
                dataframe.drop([column], axis=1, inplace=True)
        return dataframe

    @staticmethod
    def remove_records(dataframe):
        """
        Removes the rows where if a column has 20 % to 30 % of missing data.

        :param      dataframe:  The dataframe
        :type       dataframe:  { pandas.dataframe }
        :return     dataframe:  Dataframe with the rows dropped
        
        Weblink:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4

        Usage:
        ======
        >>Check.remove_records(data)
        >>dataframe

        """
        missing_dict = Check.percentage_missing(dataframe)
        for column in missing_dict:
            if missing_dict[column] >= 20.00 and missing_dict[column] < 30.00:
                dataframe = dataframe[dataframe[column].isnull() != bool(True)]
        return dataframe


    @staticmethod
    def encoding_categorical(column):
        """
        Gives the encoding of a categorical column.

        :param      col:  column name
        :type       col:  { pandas.series | list | tuple }
        :return     list: List of updated column and dict: dictionary of encoded values

        web:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4


        Usage:
        ======
            >> Check.encoding_categorical(data['Age'])
            >> False
        """
        encoded = column.astype('category').cat.codes
        encoded_col = list(encoded)
        describe_encoding = pd.Series(column, index=encoded_col).to_dict()
        return encoded_col, describe_encoding


    @staticmethod
    def replace_mean(column):
        """
        Replaces the missing values of a column with its mean.
        :param       column:  The column
        :type        column:  { pandas.series | list | tuple }
        :return      column:  Updated column after replacing missing values with mean


        Weblink:https://towardsdatascience.com/how-to-handle-missing-data-8646b18db0d4

        Usage:
        ======
        >> Check.replace_mean(data['population'])
        >> series
        """
        try:
            return column.fillna(column.mean())
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def replace_mode(column):
        """
        Replaces the missing values of a column with its mode.
        :param       column:  The column
        :type        column:  { pandas.series | list | tuple }
        :return      column:  Updated column after replacing missing values with mean

        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/

        Usage:
        ======
        >> Check.replace_mode(data['population'])
        >> series
        """
        try:
            return column.fillna(column.mode()[0])
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def replace_median(column):
        """
        Replaces the missing values of a column with its median.
        :param       column:  The column
        :type        column:  { pandas.series | list | tuple }
        :return      column:  Updated column after replacing missing values with median

        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/
         
        Usage:
        ======
        >> Check.replace_median(data['population'])
        >> series
        """
        try:
            return column.fillna(column.median())
        except AttributeError:
            print("Method only supported pandas.cores.series")


    @staticmethod
    def is_date(date):
        """
        Determines whether the specified `x` is date.

        :param      x:    string of date type
        :type       x:    string
        :returns    True | False

        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/
        Weblink:
        Usage:
        ======
            >> Check.is_date("20-20-2020")
            >> True

        """
        date_patterns = ["%m-%d-%Y", "%d-%m-%Y", "%Y-%m-%d"]
        for pattern in date_patterns:
            try:
                return datetime.datetime.strptime(date, pattern).date()
            except ValueError:
                return False

    @staticmethod
    def is_datetime(date):
        """
        Determines whether the specified `x` is datetime.

        :param      x:    { parameter_description }
        :type       x:    { type_description }
        :returns    True | False

        weblink:https://www.statisticshowto.com/probability-and-statistics/types-of-variables/

        Usage:
        ======
            >> Check.is_datetime("2020-02-20 00:00:00")
            >> True
            >> Check.is_datetime("2020-02-01")
            >> False

        """
        matches = datefinder.find_dates(date)
        for match in matches:
            if(match.hour > 0 and match.minute > 0 and match.second > 0):
                return True

        return False



# -*- coding: utf-8 -*-
import os
import six
import shutil
import yaml
import click
import json
import pickle
import logging
import pandas as pd
import numpy as np

from pyfiglet import figlet_format, Figlet
from prettytable import PrettyTable
from cognito import *
from cognito.table import Table
from cognito.story import get_interesting_stories
from datetime import datetime
from tornado import template
from tqdm import tqdm, trange



VERSION = 'version 0.0.1-beta-1'


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpEncoder, self).default(obj)

def read_yaml(filename):
    ''' take filename as parameter and convert yaml to ordereddict '''
    return yaml.load(open(filename))


custom_fig = Figlet(font='slant')
click.echo(custom_fig.renderText('cognito'))
click.echo(VERSION)



@click.group()
@click.option('--version', is_flag=True, default=False)
def cli(version):
    '''  
        Generate ML consumable datasets using advanced data preprocessing
        and data wrangling.        

        USAGE: \n
        $ cognito prepare -m ml --input filepath --out filepath 

    '''
    pass


@cli.command('report')
def report():

    preformat = """
        <link rel="stylesheet" type="text/css" href="https://cdnjs.cloudflare.com/ajax/libs/bulma/0.8.2/css/bulma.min.css">
        <img src="https://cognito.readthedocs.io/en/latest/_images/logo.png" width="200px"/>
        <h3>Summary Report</h3>
        <button class="button is-success is-rounded" onclick="window.print()">Generate PDF</button>
    """

    file_loader = FileSystemLoader(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))
    env = Environment(loader=file_loader)
    template = env.get_template('index.html')
    print(template.render(data=preformat))
    

@cli.command('audit', short_help=": Audit the given dataset and generate report")
@click.option('--inp', '-i', help="Input dataset file in following format .csv", required=True, metavar='<path>')
@click.option('--save', '-o', help="Enter the filename to save html", metavar='<path>')
def audit(inp, save):
    
    if inp:
        start_time = datetime.now()
        description = PrettyTable(['Name', 'Values'])
        description.align['Name'] = "l"
        

        table = PrettyTable([
            'Features', 
            'Type', 
            'Value Type', 
            'Outliers', 
            'Missing', 
            '(%) Missing',
            'Distinct Count',
            'Min',
            'Mean',
            'Max',
            'Zeros',
            '(%) Zeros',
            'Memory Size'            
        ])
        try:
            df_raw = Table(inp)
            df = df_raw.data
            features = df.columns
            table.align["Features"] = "l"
            table.align["Value Type"] = "l"
            table.sortby = "Distinct Count"
            table.reversesort = True

            # Generate dynamic analytical stories
            stories = get_interesting_stories(df_raw)


            # Group duplicates
            dups = df.groupby(df.columns.tolist()).size().reset_index().rename(columns={0:'count'})

            description.add_row(['Total variables', df.shape[1]])
            description.add_row(['Total Observations', df.shape[0]])
            description.add_row(['Missing Cells', df.isnull().sum().sum()])
            description.add_row(['(%) Missing Cells', df.isnull().sum().sum() / len(df)])
            description.add_row(['Duplicate Rows', dups['count'].sum() - dups.shape[0]])
            description.add_row(['(%) Duplicate Rows', (dups['count'].sum() - dups.shape[0]) / len(dups)])
            description.add_row(['Total Size of Memory', str(df.memory_usage().sum() / 1000) + 'KiB'])
            description.add_row(['🍋 Total Categorical', count_categorical(df)])
            description.add_row(['🔟 Total Continuous', count_continuous(df)])
            description.add_row(['Started at', start_time.strftime("%d-%b-%y %H:%M:%S")])

            for col in tqdm(features, ascii=True, desc="Auditing.. : "):
                table.add_row([
                    col.strip(),
                    df[col].dtypes,
                    type_of_variable(df[col]), 
                    check_outlier(df[col]),
                    check_missing(df[col]), 
                    column_missing_percentage(df[col]), 
                    distinct_count(df[col]),
                    count_min(df[col]),
                    count_mean(df[col]),
                    count_max(df[col]),
                    df[col].isin([0]).sum(),
                    round(df[col].isin([0]).sum() / len(df.columns), 2),
                    str(df[col].memory_usage() / 1000) + 'KiB'
                ])

            end_time = datetime.now()
            description.add_row(['Ended at', end_time.strftime("%d-%b-%y %H:%M:%S")])
            description.add_row(['Time Elapsed', (end_time - start_time).total_seconds()])



            # Save the report into HTML
            if save:
                
                desc_html = description.get_html_string()
                description_summary = desc_html.replace('<table>', '<table class="table is-bordered">')


                report_html = table.get_html_string()
                report_html = report_html.replace('<table>', '<table class="table is-bordered">')

                
                stories_json = [{'question': row['question'], 'answer': row['answer']} for row in stories]
                loader = template.Loader(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))

                total_categorical = list(df_raw.get_categorical())
                total_continuous = list(df_raw.get_numerical())

                open(save + '.html', 'wb').write(loader.load("index.html").generate(**locals()))
                click.echo('Report generated with name ' + save + '.html')


            click.echo(description)
            click.echo(table)

        except FileNotFoundError as e:
            logging.warning("Given input file doesn't exists")
            print(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'templates'))
       



@cli.command('inverse', short_help=": Inverse re-transform generated dataset")
@click.option('--report', '-r', prompt='Select any one operation', type=click.Choice(['transform', 'report']))
def inverse(report):
    """ Inverse transform generated Machine Learning friendly dataset """

    if report == 'report':
        inverse_transform_report()
    elif report == 'transform':
        inverse_transform()


@cli.command('prepare', short_help=': Transform given dataset')
@click.option('--mode', '-m', type=click.Choice(['ml', 'decode', 'autoML', 'help', 'report'], \
    case_sensitive=False), help="Set any mode such as `prepare`, `autoML`, `clean`", metavar='<path>')
@click.option('--inp', '-i', help="Input dataset file in following format .csv", required=True, metavar='<path>')
@click.option('--out', '-o', help="Select output desitnation", required=True, metavar='<path>')
def prepare(mode, inp, out):
    """ Transform the given dataset file """

    if mode == 'help':
        # log("Cognito CLI", color="blue", figlet=True)
        click.echo(custom_fig.renderText('cognito'))


    if mode == 'ml':

        df = Table(inp)
        response, encoder = df.generate()
        click.echo(save_to(out, response, encoder))


    if mode == 'report':
        
        df = Table(inp)
        table = PrettyTable(['Features', 'Feature Type', 'Outliers', '% of outliers', 'Missing', '%of missing'])
        for col in df.columns():
            table.add_row([col, '', '', '', '', ''])
        click.echo(table)   



if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from __future__ import print_function
import os, glob
import sys
import math
import re
import click
import pickle
from datetime import datetime
from os import path
import pandas as pd
import numpy as np
import datefinder
from cognito.table import Table
from tqdm import tqdm
from prettytable import PrettyTable




def is_working(column="Cognito!"):
    """
    Determines whether the specified command is working.
    :param      column:  The column
    :type       column:  string
    """
    return "Hello, %s! How are you %s?"%(column, column)


def is_categorical(column):
    """
    Determines whether the specified col is categorical.

    :param      col:  column name
    :type       col:  { pandas.series | list | tuple }
    :return     boolean: True | False

    Usage:
    ======
        >> is_categorical(data['Age'])
        >> True
    """
    try:
        return bool(True) if column.dtypes == 'object' else bool(False)

    except AttributeError:

        print("Method only supported pandas.cores.series")


def is_continuous(column):
    """
    Determines whether the specified col is continuous.

    :param      col:  column name
    :type       col:  { pandas.series | list | tuple }
    :return     boolean: True | False

    Usage:
    ======
        >> is_continuous(data['Age'])
        >> False
    """
    try:
        return bool(True) if column.dtypes == 'float64' else bool(False)

    except AttributeError:

        print("Method only supported pandas.cores.series")


def is_discrete(column):
    """
    Determines whether the specified column is discrete.

    :param      column:  column name
    :type       column:  { pandas.series | list | tuple }
    :return     boolean: True | False

    Usage:
    ======
        >> is_discrete(data['Age'])
        >> True
    """
    try:
        return bool(True) if column.dtypes == 'int64' else bool(False)

    except AttributeError:

        print("Method only supported pandas.cores.series")


def is_identifier(column):
    """
    Determines whether the specified column is identifier.

    :param      column:  The column
    :type       column:  { pandas.series | list | tuple }
    :return     boolean: True | False

    Usage:
    ======
        >> is_identifier(data['Age'])
        >> True

    """
    try:
        return bool(True) if column.nunique() == column.shape[0] else bool(False)

    except AttributeError:

        print("Method only supported pandas.cores.series")


def is_missing(column):
    """
    Determines whether the specified column has missing values.

    :param      column:  The column
    :type       column:  { pandas.series | list | tuple }
    :return     boolean: True | False

    Usage:
    ======
        >> is_missing(data['population'])
        >> True

    """
    try:
        return bool(True) if column.isnull().values.any() == bool(True) else bool(False)

    except AttributeError:

        print("Method only supported pandas.cores.series")


def ignore_identifier(dataframe):
    """
    Drops the table if the column is an identifier.

    :param      column:  The column
    :type       column:  { pandas.series | list | tuple }
    :return     DataFrame:Updated DataFrame

    Usage:
    ======
        >> ignore_identifier(data)
        >> Updated Dataframe
    """
    col = list(dataframe)
    for i in col:
        if is_identifier(dataframe[i]):
            dataframe.drop([i], axis=1, inplace=True)
    return dataframe


def is_outlier(column, threshold):
    """
    Returns all the outliers in the column.

    :param      column:   The column
    :type       column:   {pandas.series | list | tuple}
    :return     list:     List of all the ouliers in the column

    Usage:
    =====
       >> is_outlier(data['population'])
       >> [6815.0, 6860.0, 11551.0]
    """
    column = pd.Series(column)
    outliers = []
    mean = np.mean(column)
    std_dev = np.std(column)
    for value in column:
        z_score = (value - mean) / std_dev
        if np.abs(z_score) > threshold:
            outliers.append(value)
    return outliers


def column_missing_percentage(col):
    """
    Column missing percentage
    
    :param      col:  The col
    :type       col:  { type_description }
    
    :returns:   { description_of_the_return_value }
    :rtype:     { return_type_description }
    """
    return col.isnull().sum() * 100 / len(col)


def percentage_missing(dataframe):
    """
    Calculates the percentage of missing value in each column of dataframe.

    :param       dataframe:  The dataframe
    :type        dataframe:  { pandas.dataframe }
    :return      dictionary:  Dictionary of column name with percentage of missing values

    Usage:
    ======
    >> percentage_missing(data)
    >> {Price:0.00, Age:10.00}

    """
    missing = dataframe.isnull().sum()
    row_size = len(dataframe.index)
    missing_dict = {}
    for i in list(dataframe.columns):
        perc = round(missing[i] / row_size * 100.0, 2)
        missing_dict.update({i:perc})
    return missing_dict



def distinct_count(col):
    """
    Count the distinct values
    
    :param      col:  The col
    :type       col:  { type_description }
    
    :returns:   { description_of_the_return_value }
    :rtype:     { return_type_description }
    """
    return len(col.value_counts())


def count_min(col):

    if is_continuous(col):
        return col.min()
    else:
        return "🍋"

def count_max(col):

    if is_continuous(col):
        return col.max()
    else:
        return "🍋"

def count_mean(col):

    if is_continuous(col):
        return round(col.mean(), 2)
    else:
        return "🍋"


def count_categorical(df):
    count = 0
    for col in df.columns:
        if is_categorical(df[col]):
            count = count + 1
    return count

def count_continuous(df):
    count = 0
    for col in df.columns:
        if is_continuous(df[col]):
            count = count + 1
    return count



def remove_columns(dataframe):
    """
    Removes the column containing 60% or more missing data.

    :param      dataframe:  The dataframe
    :type       dataframe:  { pandas.dataframe }
    :return     dataframe:  Dataframe with the columns dropped

    Usage:
    ======
    >>remove_columns(data)
    >>dataframe

    """
    missing_dict = percentage_missing(dataframe)
    for column in missing_dict:
        if missing_dict[column] >= 60.00:
            dataframe.drop([column], axis=1, inplace=True)
    return dataframe

def remove_records(dataframe):
    """
    Removes the rows where if a column has 20 % to 30 % of missing data.

    :param      dataframe:  The dataframe
    :type       dataframe:  { pandas.dataframe }
    :return     dataframe:  Dataframe with the rows dropped

    Usage:
    ======
    >>remove_records(data)
    >>dataframe

    """
    missing_dict = percentage_missing(dataframe)
    for column in missing_dict:
        if missing_dict[column] >= 20.00 and missing_dict[column] < 30.00:
            dataframe = dataframe[dataframe[column].isnull() != bool(True)]
    return dataframe


def encoding_categorical(column):
    """
    Gives the encoding of a categorical column.

    :param      col:  column name
    :type       col:  { pandas.series | list | tuple }
    :return     list: List of updated column and dict: dictionary of encoded values
    Usage:
    ======
        >> encoding_categorical(data['Age'])
        >> False
    """
    encoded = column.astype('category').cat.codes
    encoded_col = list(encoded)
    describe_encoding = pd.Series(column, index=encoded_col).to_dict()
    return encoded_col, describe_encoding


def replace_mean(column):
    """
    Replaces the missing values of a column with its mean.

    :param       column:  The column
    :type        column:  { pandas.series | list | tuple }
    :return      column:  Updated column after replacing missing values with mean

    Usage:
    ======
    >> replace_mean(data['population'])
    >> series
    """
    try:
        return column.fillna(column.mean())

    except AttributeError:

        print("Method only supported pandas.cores.series")


def replace_mode(column):
    """
    Replaces the missing values of a column with its mode.

    :param       column:  The column
    :type        column:  { pandas.series | list | tuple }
    :return      column:  Updated column after replacing missing values with mean

    Usage:
    ======
    >> replace_mode(data['population'])
    >> series
    """
    try:
        return column.fillna(column.mode()[0])

    except AttributeError:

        print("Method only supported pandas.cores.series")


def replace_median(column):
    """
    Replaces the missing values of a column with its median.

    :param       column:  The column
    :type        column:  { pandas.series | list | tuple }
    :return      column:  Updated column after replacing missing values with median

    Usage:
    ======
    >> replace_median(data['population'])
    >> series
    """
    try:
        return column.fillna(column.median())

    except AttributeError:

        print("Method only supported pandas.cores.series")


def is_date(date):
    """
    Determines whether the specified `x` is date.

    :param      x:    string of date type
    :type       x:    string
    :returns    True | False
    Usage:
    ======
        >> is_date("20-20-2020")
        >> True

    """
    date_patterns = ["%m-%d-%Y", "%d-%m-%Y", "%Y-%m-%d"]
    for pattern in date_patterns:
        try:
            return datetime.datetime.strptime(date, pattern).date()
        except ValueError:
            return False

def is_datetime(date):
    """
    Determines whether the specified `x` is datetime.

    :param      x:    { parameter_description }
    :type       x:    { type_description }
    :returns    True | False

    Usage:
    ======
        >> is_datetime("2020-02-20 00:00:00")
        >> True
        >> is_datetime("2020-02-01")
        >> False

    """
    matches = datefinder.find_dates(date)
    for match in matches:
        if(match.hour > 0 and match.minute > 0 and match.second > 0):
            return True

    return False


def col_missing_percentage(df, col):
    """
    Column wise missing value percentagle
    
    :param      df:   { parameter_description }
    :type       df:   { type_description }
    :param      col:  The col
    :type       col:  { type_description }
    """
    return df[col].isnull().sum() * 100 / len(df)


def type_of_variable(col):
    """
    Get type of variable
    
    :param      col:  The col
    :type       col:  { type_description }
    """
    if is_categorical(col):
        return 'categorical'

    elif is_continuous(col):
        return 'numerical'

    elif is_discrete(col):
        return 'discrete'

    elif is_identifier(col):
        return 'cardinal'

    elif is_datetime(col):
        return 'datetime'

    else:
        return col.dtypes


def check_outlier(col):
    """
    Check the outlier for the given column
    
    :param      col:  The col
    :type       col:  { type_description }
    """
    if is_continuous(col):
        if is_outlier(col, 3):
            # return "❌"
            return "🚩"
        else:
            # return "☘"
            return "✅"
          
    else:
        return "🍋"


def check_missing(col):

    if is_missing(col):
        return "🔴"
    else:
        # return "☘"
        return "✅"



def list_diff(l1, l2):
	"""
	Get the difference between two list `l1` and `l2`
	
	:param      l1:   The l 1
	:type       l1:   { type_description }
	:param      l2:   The l 2
	:type       l2:   { type_description }
	
	:returns:   { description_of_the_return_value }
	:rtype:     { return_type_description }
	"""
	return (list(set(l1) - set(l2))) 


def get_all_files():
    files = [f for f in os.listdir('.') if os.path.isfile(f)]
    return files


def save_to(path, df, encoder):
    """
    Save the encoded dataframe to csv file and 
    picle file.
    
    :param      path:     The path
    :type       path:     { type_description }
    :param      df:       { parameter_description }
    :type       df:       { type_description }
    :param      encoder:  The encoder
    :type       encoder:  { type_description }
    """
    filename = os.path.basename(path)

    if '.' in filename:
        fname, ext = filename.split('.')
    else:
        fname = filename

    path = os.path.dirname(path)
    save_path = os.path.join(path, fname)

    # make directory
    try:
        os.mkdir(save_path)

        #filenames
        pkl_file = os.path.join(save_path, filename + '.pkl')

        if '.csv' not in filename:
            filename = filename + '.csv'    

        df_file = os.path.join(save_path, filename)

        df.to_csv(df_file, index=False)
        f = open(pkl_file, "wb")
        pickle.dump(encoder, f)
        f.close()
        return df

    except Exception as e:

        click.echo(
            click.style(
                "Abort: The {} file already exists.", fg="red"
            ).format(os.path.join(save_path, filename)), err=True)



def inverse_transform():
    """
    Inverse transform pickled encoders and output dataframe to 
    its original form.
    """
    pickles = glob.glob("*.pkl")
    datasets = glob.glob("*.csv")

    if len(pickles) > 0 and len(datasets) > 0:
        
        # pick the first pickle 
        first_pickle = pickles.pop(0)

        # Get the file related the given pickle
        fname = re.sub("\.csv\..*","", first_pickle)

        # Grab the encoder from pickle
        encoder = pickle.load(open(first_pickle, 'rb'))

        fname, ext  = first_pickle.split('.')
        
        # Load the given dataset into memory
        df = Table(fname + '.csv')

        columns = encoder.keys()

        for col in tqdm(columns, ascii=True, desc="Decoding all : "):
            # Pick the encoder
            le = encoder[col]
            df.data[col] = le.inverse_transform(df.data[col])

        output_file =  fname + '_inversed_' + datetime.now().strftime("%d-%B-%Y-%H-%M-%S") + '.csv'
        df.data.to_csv(output_file, index=False)
        print(output_file)
            
    else:
        click.echo('No cognito decoded files found')



def inverse_transform_report():
    """
    Inverse transform pickled encoders and output dataframe to 
    its original form.
    """
    pickles = glob.glob("*.pkl")
    datasets = glob.glob("*.csv")

    if len(pickles) > 0 and len(datasets) > 0:
        
        # pick the first pickle 
        first_pickle = pickles.pop(0)

        # Get the file related the given pickle
        fname = re.sub("\.csv\..*","", first_pickle)

        # Grab the encoder from pickle
        encoder = pickle.load(open(first_pickle, 'rb'))
        columns = encoder.keys()

        # Load the given dataset into memory
        df = Table(fname + '.csv')
        x = PrettyTable()
        x.field_names = ["Feature Name", "% Missing", "Outliers"]
        x.align["Feature Name"] = "l"
       

        for col in tqdm(columns, ascii=True, desc="Decoding all : "):
            le = encoder[col]
            # print(le.inverse_transform(df.data[col].unique))
            # click.echo(col)
            missing = col_missing_percentage(df.data, col)
            x.add_row([col, missing, ''])

        print(x)






import os
import sys
import pandas as pd

class Transform:

    def __init__(self):
		pass

	def split_date(x):
		"""
		Split the given string `x` into `day,month,year`
		
		:parameter_description x:	date
		:type_description x:	string
		:return		return Dict
		
		Usage:
		======
			>> T = Transform()
			>> T.__split_date('2020-10-18')
			>> {'day': 18, 'month': 10, 'year': 2020 }

		"""
		import datetime
		li = x.split('-')
		date = ("-").join(li)
		d = datetime.datetime.strptime(date, "%d %b %Y")
		list1 = [d.day,d.month,d.year]
		list2 = ['day','month','year']
		res = {list2[i]: list1[i] for i in range(len(list2))} 
		return res

	@staticmethod
	def split_dates(x):
		"""
		Splits dates given as `pandas.core.series` into multple
		columns consist of `day, month, year` using the above
		`__split_date` private method.
		
		:param      x:    { parameter_description }
		:type       x:    { type_description }
		return 		pandas.core.series

		Usage:
		======
			>> Transform.split_dates([2020-10-02, 2019-10-01, 2020-11-11])
			>> day, month, year
			    02     10   2020
			    01     10   2019
			    11     11   2020

		"""
		pass


	@staticmethod
	def numerical(column):
		"""
		Given a `pandas.core.series` consist of categorical data
		encode it into numerical mapping.
		
		:param      column:  list of categorical values
		:type       column:  pandas.core.series
		:return     list of dictionary consisting of key: value

		Usage:
		======
			>> T = Transform()
			>> T.numerical(['cat', 'rat', 'mice', 'mice', 'rat'])
			>> [
				'cat':  0,
				'rat':  1,
				'mice': 3
				]
		"""
		encoded = column.astype('category').cat.codes
		encoded_col = list(encoded)
		describe_encoding = pd.Series(encoded_col, index=column).to_dict()
		return describe_encoding



"""
Cognito module
"""
from codecs import open as codecs_open
from setuptools import setup, find_packages



# Get the long description from the relevant file
with codecs_open('README.rst', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


REQUIRES = [
    'tqdm',
    'numpy',            # REQ: vector algebra operations
    'scipy',
    'numpy',
    'click',            # REQ: command line interfacing
    'pandas',           # REQ: (conda) sparx.data.filter()
    'tornado',          # REQ: report generation engine
    'PyYAML',           # REQ: configuration management
    'pyfiglet',         # REQ: better cli interface  
    'datefinder',     
    'PrettyTable',
    'sparklines',
    'scikit-learn'
]


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    "Operating System :: OS Independent",
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python',
    'Topic :: Internet :: WWW/HTTP',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Utilities',
    'Topic :: Scientific/Engineering'

]


DOWNLOAD_URL = ""
PROJECT_URLS = {
    "Bug Tracker": "https://github.com/Cleverinsight/congito/issues",
    "Documentation": "https://cognito.readthedocs.io/en/latest/",
    "Source Code": "https://github.com/Cleverinsight/congito",
}


setup(name='cognito',
      version='0.0.1b1',
      description=u"Auto ML Dataset Transformer & Auto Data Storyteller",
      long_description=LONG_DESCRIPTION,
      classifiers=CLASSIFIERS,
      keywords=['Automated Data Storyteller', 'Data Wrangler', 'Data Preprocessing',\
       'Machine Learning', 'Hot Encoder', 'Outlier Detection'],
      author=u"Bastin Robins .J",
      author_email='robin@cleverinsight.co',
      url='https://github.com/cleverinsight',
      download_url='https://github.com/CleverInsight/cognito/releases',
      project_urls=PROJECT_URLS,
      license='BSD',
      packages=[pkg for pkg in find_packages() if not pkg.startswith('test')],
      include_package_data=True,
      zip_safe=False,
      install_requires=REQUIRES,

      extras_require={
          'test': ['pytest'],
      },

      entry_points="""
      [console_scripts]
      cognito=cognito.scripts.cli:cli
      """)



#pylint: disable = E0202
#pylint: disable = R0904
#pylint: disable = C0200
#pylint: disable = E0602
#pylint: disable = C0103
'''
Importing all the libraries needed
'''
import pickle
import math
from collections import Counter
import pandas as pd
import numpy as np
from scipy.stats.stats import kendalltau
from scipy.stats import pointbiserialr
from sklearn.preprocessing import LabelEncoder
from sklearn import preprocessing
from tqdm import tqdm


def list_diff(l1, l2):
    """
    Get the difference between two list `l1` and `l2`
    
    :param      l1:   The l 1
    :type       l1:   { type_description }
    :param      l2:   The l 2
    :type       l2:   { type_description }
    
    :returns:   { description_of_the_return_value }
    :rtype:     { return_type_description }
    """
    return (list(set(l1) - set(l2))) 

class Table:
    """
    Table takes a csv file as input and converts to
    a datframe to perform specific operations.
    """
    def __init__(self, filename):
        self.data = pd.read_csv(filename)


    def data(self):
        """
        Returns the actual dataframe from taken
        form processing.
        """
        return self.data


    def columns(self):
        """
        Get all the column names from the given
        dataframe `self.data`
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.columns()
        """
        return self.data.columns


    def total_columns(self):
        """
        Get the count of all column in the given
        dataframe `self.data`
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.total_columns()
        """
        return len(self.columns())


    def total_rows(self):
        """
        Get total count of rows from the current
        dataframe `self.data`.
        returns: dataframe
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.total_rows()
        """
        return self.data.shape[0]


    def get_categorical(self):
        """
        Gets the categorical columns from the given
        dataframe `self.data`
        returns: dataframe
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.get_categorical()
        """
        dataframe = pd.DataFrame()
        for i in self.data:
            if self.data[i].dtypes == 'object':
                dataframe[i] = pd.Series(self.data[i])
        return dataframe


    def get_numerical(self):
        """
        Gets the numerical columns from the given
        dataframe `self.data`
        returns: dataframe
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.get_numerical()
        """
        dataframe = pd.DataFrame()
        for i in self.data:
            if np.issubdtype(self.data[i].dtype, np.number):
                dataframe[i] = pd.Series(self.data[i])
        return dataframe


    def odd_rows(self):
        """
        Get all odd indexed counted rows from the given
        dataframe `self.data`
        returns: dataframe
        Usage:
        ======
            >>> data = Table('filename.csv')
            >>> data.odd_rows()
        """
        return self.data.iloc[1::2]


    def even_rows(self):
        """
        Get all even indexed counted rows from the given
        dataframe `self.data`
        returns: dataframe
        """
        return self.data.iloc[:-2:2]


    #def apply(self):


    def summary(self):
        """
        Return the dataframe descriptive statistics
        returns: dataframe summary
        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.summary()
        """
        return self.data.describe()


    def hot_encoder_categorical(self, column):
        """
        Returns the pandas.series with hashtable in Dict structures
        returns: pandas.series, dict
        Usage:
        ======
            >>> df.hot_encoder_categorical(col_name)
        """
        one_hot = pd.get_dummies(self.data[column])
        return one_hot


    def convert_to_bin(self, column):
        """
        Returns the columns with more than 50% threshold to
        newly created bin pandas.series
        returns: pandas.series
        descriptions: list of newly created bin values
        :param      column
        :type       name of the column
        :returns:   list of generated bins
        :rtype:     list
<<<<<<< HEAD

        Weblink: https://www.geeksforgeeks.org/binning-in-data-mining/


=======
>>>>>>> a0dc96e323316a85c3b913517ea5819103ce1ed5
        Usage:
        ======
            >>> self.convert_to_bin(col_name)
        """
        length = len(self.data[column])
        sqr = round(math.sqrt(length))
        maximum = int(max(self.data[column]))
        minimum = int(min(self.data[column]))
        bin_size = int(round((maximum - minimum) / sqr))
        quantity = round(maximum/bin_size)
        bins = []
        for low in range(int(minimum - 1), int(minimum + quantity * bin_size + 1), bin_size):
            bins.append((low+1, low + bin_size))
        return bins


    def correlation(self, mode="pearson"):
        """
        Return the pairwise correlation of the given
        dataframe `self.data` and return dataframe with
        respective dataframe.
        :param      mode:  `Pearson`, `Kendall`, `Spearman`, `Point-Biserial`
        :type       mode:  string
        :returns:   correlation matrix
        :rtype:     dataframe

        Weblink: https://www.geeksforgeeks.org/mathematics-covariance-and-correlation/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.correlation()
        """
        if mode == "pearson":
            pearsoncorr = self.data.corr(method='pearson')
        elif mode == "kendall":
            pearsoncorr = self.data.corr(method='kendall')
        elif mode == "spearman":
            pearsoncorr = self.data.corr(method='spearman')
        return pearsoncorr


    def covariance(self):
        """
        Return the covariance of the given
        dataframe `self.data` and return dataframe with
        respective dataframe.
<<<<<<< HEAD
        Weblink: https://www.geeksforgeeks.org/mathematics-covariance-and-correlation/
        
=======
        Ref: https://www.theanalysisfactor.com/covariance-matrices/
>>>>>>> a0dc96e323316a85c3b913517ea5819103ce1ed5
        returns :  dataframe
        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.covariance()
        """
        result = self.data.cov()
        return result


    def slice(self, columns):
        """
        Return only the columns sliced
        from `self.data` based on given `columns` parameter
        :param      columns:  list or tuple
        :type       columns:  { list of columns name  }
        returns: dataframe of only given column names.

        weblink:https://www.geeksforgeeks.org/python-slice-function/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.slice(['name', 'age'])
            ---------------
              name  | age
            ---------------
              Sam   | 20
            ---------------
              Jack  | 40
            ---------------
        """
        return self.data[columns]


    def binning(self, column, bins):
        """
        Return dataframe of select column convert into bins as given
        parameter
        :param      col:   The column
        :type       col:   { column name to be selected }
        :param      bins:  The bins
        :type       bins:  list of  of bins to convert
<<<<<<< HEAD

        weblink: https://www.geeksforgeeks.org/binning-in-data-mining/

=======
        :param      bins:  The bins
        :type       bins:  list of  of bins to convert
>>>>>>> a0dc96e323316a85c3b913517ea5819103ce1ed5
        returns: dataframe of given column as bins
        example:
            # Numerical Binning Example

                Value      Bin
                0-30   ->  (0-30)
                31-70  ->  (31-70)
                71-100 ->  (71-100)


            # Categorical Binning Example
                Value      Bin
                Spain  ->  Europe
                Italy  ->  Europe
                Chile  ->  South America
                Brazil ->  South America
        Usage:
            >>> df = Table('filename.csv')
            >>> df.binning('age', bins=[], labels=['low', 'med', 'high'])
            ------------------
            |  age  | age_bin
            ------------------
            |  14   | low
            ------------------
            |  90   | high
            ------------------
            |  14   | low
            ------------------
            |  40   | middle
            ------------------
            |  14   | low
            ------------------
        """
        binned_weight = []
        data = pd.DataFrame(columns=['Value', 'Bin'])
        for val in self.data[column]:
            for i in range(0, len(bins)):
                if bins[i][0] <= val < bins[i][1]:
                    data = data.append({'Value':val, 'Bin':bins[i]}, ignore_index=True)
                    binned_weight.append(i)
        freq = Counter(binned_weight)
        return (data, freq)


    def fix_outlier_with_std_deviation(self, column):
        """
        Take the column from `self.data` and check for outlier
        fix the outlier using mode : std deviation
        Ref: https://towardsdatascience.com/feature-engineering-for-machine-learning-3a5e293a5114
        :param      column:  The column
        :type       column:  { column name as string }
        :param      mode:  mode like std deviation
        :type       column:  { string }
        returns: dataframe without outlier

        Weblink: https://www.kdnuggets.com/2017/02/removing-outliers-standard-deviation-python.html

        Usage:
        ======
            >>> df = Table('filename.csv')

            >>> df.fix_outlier_with_std_deviation('age', 3)
        """
        outliers = []
        mean = np.mean(self.data[column])
        std_dev = np.std(self.data[column])
        for value in self.data[column]:
            z_score = (value - mean) / std_dev
            if np.abs(z_score) > 3:
                outliers.append(value)
        for i, _ in enumerate(self.data[column]):
            if i in outliers:
                i.replace(std_dev, inplace=True)
        return self.data[column]


    def fix_missing(self, column):
        """
        Take the column from `self.data` and fix the type of columns
        and clean the missing values based on categorical or numerical
        :param      column:  The column
        :type       column:  { column name }
        returns: dataframe | pandas.core.series

        Weblink: https://www.geeksforgeeks.org/working-with-missing-data-in-pandas/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.fix_missing('age')
        """
        self.ignore_cardinal()
        continuous = self.get_numerical().columns.tolist()
        categorical = self.get_categorical().columns.tolist()
        if column in continuous:
            self.data[column].fillna(self.data[column].mean(), inplace=True)
        elif column in categorical:
            self.data[column].fillna(self.data[column].mode()[0], inplace=True)
        return self.data[column]


    def imputer(self, column, value):
        """
        Take the columns from `self.data` and fill the missing NA
        values with given value parameter based on the type of column
        :param      column:  The column
        :type       column:  { column name }
        :param      value:  The value by which NA will be replaced
        :type       value:  { string, number, boolean }
        returns: dataframe | pandas.core.series

        Weblink: https://www.geeksforgeeks.org/working-with-missing-data-in-pandas/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.imputer('age', 10)
        """
        self.data[column].fillna(value, inplace=True)
        return self.data[column]


    def ignore_cardinal(self):
        """
        Take the dataframe `self.data` and remove all cardinality columns
        and return the dataframe
        returns: dataframe without cardinality columns

        Weblink:https://www.geeksforgeeks.org/python-named-entity-recognition-ner-using-spacy/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.ignore_cardinal()
        """
        data = self.data.copy()
        for i in data:
            if len(data[i]) == data[i].nunique(dropna=True):
                data.drop(i, axis=1, inplace=True)
        return data


    def encode_column(self, column):
        """
        Encodes a column using the numerical values and
        return the dictionary for mapping.
        :param      column:  The column
        :type       column:  { column name }
        returns: dataframe, mapper dictionary

        Weblink: https://www.geeksforgeeks.org/ml-label-encoding-of-datasets-in-python/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.encode_column('Country')
            >>> (dataframe, {'0': 'US', '1': 'India', '2': 'Europe'})
        """
        encode = LabelEncoder()
        data = encode.fit_transform(self.data[column])
        return data, encode


    def list_cardinal(self):
        """
        Return the list of all cardinality columns from the given
        `self.data`
        :returns:   { list of all cardinality values }
        :rtype:     { List }
        """
        return [col for col in self.data if len(self.data[col]) == self.data[col].nunique()]


    def generate(self):
        """
        No docstring for the time being
        """
        cardinal_col = self.list_cardinal()
        categorical_col = list_diff(self.get_categorical().columns, cardinal_col)
        numerical_col = list_diff(self.get_numerical().columns, cardinal_col)

        # Fix categorical and numerical columns
        for col in tqdm(categorical_col + numerical_col, ascii=True, desc="Imputing missing : "):
            self.fix_missing(col)

        data = self.data.drop(cardinal_col, axis=1)

        # Encode categorical variables
        encoders = {}

        for col in tqdm(categorical_col, ascii=True, desc="Encoding : "):
            x, y = self.encode_column(col)
            data[col] = x
            encoders[col] = y

        return data, encoders


    def scale(self, columns, mode='minmax'):
        """
        Take a dataframe self.data and list of columns, scales the feature with distribution value
        between 0 and 1
        :param      column:  columns
        :type       column:  [list of colums to be scaled]
        :param      column:  mode
        :type       column:  mode for scaling  MinMax/std_dist
        returns: dataframe of columns after scaling

        Weblink: https://www.geeksforgeeks.org/scales-of-measurement/

        Usage:
        ======
            >>> df = Table('filename.csv')
            >>> df.scale(columns, mode)
            >>> dataframe
        """
        original_data = self.slice(columns)
        if mode == 'minmax':
            scaler = preprocessing.MinMaxScaler(feature_range=(0, 1))
            transformed_data = scaler.fit_transform(original_data)
        elif mode == 'std_dist':
            scaler = preprocessing.StandardScaler()
            transformed_data = scaler.fit_transform(original_data)
        return transformed_data


    def categorical_columns(self):
        """
        Get only the categorical columns
        
        :returns:   { description_of_the_return_value }
        :rtype:     { return_type_description }
        """
        return self.get_categorical().columns
